package czerwony.kapturek.jbwheel;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;

import com.borland.jbuilder.JBuilderMenu;
import com.borland.jbuilder.JBuilder;
import com.borland.primetime.ide.*;
import com.borland.primetime.editor.*;
import com.borland.primetime.node.*;
import com.borland.primetime.actions.*;

/** Wheel tool class  */
public class WheelTool
{
    /* notch move */
    private final static int NOTCH_MOVE = 120;

    /** configuration structure */
    private Config	conf = new Config();

    /** "not used" amount of wheel movement */
    private int		reminder;

    /** UGLY: if we cannot find anything suitable
     for scrolling but the presence of a scrollbar is found
     somewhow it can be used for scrolling. This is done mostly
     for scrolling help windows (witch for some reason do
     not have Scrollable components naither they use JScrollPane
     or JViewport).
     The only nice thing about it is that it can be used for
     switching the orientation of scrolling from default (VERTICAL)
     to HORIZONTAL is the mouse pointer is over the horizontal
     JScrollBar (in a similar way some MS programs work on Windows).
     */
    private JScrollBar	scrollFound;

    /** On JDK 1.3 trying to scroll in a context of JNI call for some reason
       (I don't have time to invastigate) freezes the whole thing.
       Because of that scrolling is delegated to additional object invoked
       using SwingUtilities.invokeLater() method.
    */
    private WheelNotifier notifier;

    /**        */
    private final static String	DLL_NAME = "jbWheel";

    /** The one and only instance of the tool */
    private static WheelTool I;

    private int action;

    public static void initOpenTool( byte major, byte minor )
    {
        I = new WheelTool();
    }

    public WheelTool()
    {
        trace( "Starting jbWheel" );
        try
        {
            if ( System.getProperty( "java.vm.version", "" ).indexOf( "1.2" )==-1 )
            {
                notifier = new WheelNotifier();
                trace( " ...running on JVM 1.3+" );
            }

            String rootDir = JBuilder.getRootDir();
            String lib = System.mapLibraryName( DLL_NAME );
            String libPath = ClassDirFinder.find( getClass().getName() );
            File libFile = new File( libPath.toString(), lib );
            if ( libFile.exists() )
                System.load( libFile.getCanonicalPath() );
            else
                System.loadLibrary( DLL_NAME );
            startCaptureWheelMessages( conf.debug );

            Icon ico = new ImageIcon( getClass().getResource("mouse.gif") );

            JBuilderMenu.GROUP_ToolsOptions.add( new AbstractAction( "WM", ico )
                {
                    public void actionPerformed( ActionEvent ev )
                    {
                        JDialog dlg = new WheelToolDialog( Browser.getActiveBrowser(), conf );
                        Rectangle r = Browser.getActiveBrowser().getBounds();
                        Dimension d = dlg.getSize();
                        r.x = (r.width - d.width) / 2;
                        r.y = (r.height - d.height) / 2;
                        r.width = d.width;
                        r.height = d.height;
                        dlg.setBounds( r );
                        dlg.show();
                    }

                    public String toString()
                    {
                        return "WheelMouse settings...";
                    }
                } );
        }
        catch ( Exception ex )
        {
            JOptionPane.showMessageDialog(null, ex.toString());
        }
    }

    /** Sets the hool for intercepting wheel signals.
     *  @return  */
    private native void startCaptureWheelMessages( boolean debug );

    /** Removes the hool for intercepting wheel signals. */
    private native void stopCaptureWheelMessages();

    public synchronized void onWheelMoved( int move, int x, int y, int mask )
    {
        trace( "wheel move notification" );
        try
        {
            if ( notifier==null )
            {
                wheelMoved( move, x, y, mask );
            }
            else
            {
                notifier.move += move;
                notifier.x = x;
                notifier.y = y;
                notifier.mask = mask;
                SwingUtilities.invokeLater( notifier );
            }
        }
        catch ( Exception ex )
        {
            trace( "E X C E P T I O N" );
            ex.printStackTrace();
        }
    }

    public void wheelMoved( int move, int x, int y, int mask )
    {
        if ( conf.reverse )
            move = -move;
        if ( move*reminder >= 0 )
            move += reminder;
        reminder = (conf.sensitivity*move)%(NOTCH_MOVE*10);
        trace( " ...amount to scroll: " + move );
        if ( move==0 )
            return;

        Integer i = (Integer)conf.actions.get( new Integer(mask) );
        if ( i==null )
            return;
        action = i.intValue();

        switch ( action )
        {
            case Config.Action.nextnode:
                try
                {
                    if ( move > 0 )
                        Browser.getActiveBrowser().nextOpenNode(true);
                    else
                        Browser.getActiveBrowser().previousOpenNode(true);
                }
                catch ( Exception ex ) {}
                break;

            default:
                actionScroll( move, x, y );
                break;
        }
    }

    private void actionScroll( int move, int x, int y )
    {
        scrollFound = null;
        JComponent c = null;

        if (conf.mode==Config.MODE_EDITORONLY)
            c = findEditorPane();
        else if (conf.mode==Config.MODE_UNDERCURSOR)
            c = findComponentAt( x, y );
        else if (conf.mode==Config.MODE_FOCUSED)
            c = findFocused();

        if ( c instanceof Scrollable )
            scrollScrollable( c, move );
        else if ( c instanceof JScrollPane )
            scrollPanel( (JScrollPane)c, move );
        else if ( c instanceof JSlider )
            scrollSlider( (JSlider)c, move );
        else if ( c instanceof JComboBox )
            scrollCombo( (JComboBox)c, move );
        else if ( scrollFound!=null )
            scrollScrollBar( move );
        else if ( conf.debug )
            trace( "nothing suitable to scroll: "
                    + (c==null ? "null" : c.getClass().getName()) );
    }


    private void scrollScrollable( JComponent comp, int move )
    {
        //System.err.println( "scrollScrollable()" );
        Rectangle r = new Rectangle();
        comp.computeVisibleRect( r );
        int i = 0;
        int orient = (scrollFound!=null && conf.followBar)
                ? scrollFound.getOrientation()
                : ( (action==Config.Action.vscroll || action==Config.Action.pgvscroll )
                    ? SwingConstants.VERTICAL : SwingConstants.HORIZONTAL );
        if ( action==Config.Action.pgvscroll || action==Config.Action.pghscroll )
        {
            i = ((Scrollable)comp).getScrollableBlockIncrement( r, orient, -move );
            if ( move<0 )
                i = -i;
        }
        else
        {
            int incr = ((Scrollable)comp).getScrollableUnitIncrement( r, orient, -move );
            i = (conf.sensitivity*move/(NOTCH_MOVE*10)) * incr;
        }

        if ( orient == SwingConstants.VERTICAL )
        {
            r.y -= i;
            if ( r.y<0 )
                r.y = Math.max( 0, r.y );
            else
                r.y = Math.min( comp.getHeight()-r.height, r.y );
        }
        else
        {
            r.x -= i;
            if ( r.x<0 )
                r.x = Math.max( 0, r.x );
            else
                r.x = Math.min( comp.getWidth()-r.width, r.x );
        }

        comp.scrollRectToVisible( r );
    }


    private void scrollPanel( JScrollPane comp, int move )
    {
        //System.err.println( "scrollPanel()" );
        JViewport vp = comp.getViewport();
        if ( vp.getView() instanceof Scrollable )
        {
            scrollScrollable( (JComponent)vp.getView(), move );
            return;
        }

        JScrollBar sb = (scrollFound!=null && conf.followBar)
            ? scrollFound
            : ( (action==Config.Action.vscroll || action==Config.Action.pgvscroll )
                ? comp.getVerticalScrollBar() : comp.getHorizontalScrollBar() );
        if ( sb!=null )
        {
            Rectangle r =  vp.getViewRect();
            int i = 0;
            if ( action==Config.Action.pgvscroll || action==Config.Action.pghscroll )
            {
                i = sb.getBlockIncrement( move>0 ? -1 : 1 );
                if ( move<0 )
                    i = -i;
            }
            else
            {
                int incr = sb.getUnitIncrement( move>0 ? -1 : 1 );
                i = (conf.sensitivity*move/(NOTCH_MOVE*10)) * incr;
            }

            if ( action==Config.Action.hscroll || action==Config.Action.pghscroll )
            {
                r.x -= i;
                if ( r.x<0 )
                    r.x = Math.max( 0, r.x );
                else
                    r.x = Math.min( comp.getWidth()-r.width, r.x );
            }
            else
            {
                r.y -= i;
                if ( r.y<0 )
                    r.y = Math.max( 0, r.y );
                else
                    r.y = Math.min( comp.getHeight()-r.height, r.y );
            }
            vp.scrollRectToVisible( r );
        }
    }


    private EditorPane findEditorPane()
    {
        Browser b = Browser.getActiveBrowser();
        Node n = b.getActiveNode();
        NodeViewer v = b.getActiveViewer( n );
        JComponent c = v.getViewerComponent();
        return findEditorPane( c );
    }


    private EditorPane findEditorPane( Component c )
    {
        Component[] cc = ((JComponent)c).getComponents();
        for ( int i=0; i<cc.length; i++ )
        {
            if ( cc[i] instanceof EditorPane )
                return (EditorPane)cc[i];
            EditorPane ep = findEditorPane( cc[i] );
            if ( ep!=null )
                return ep;
        }
        return null;
    }


    private JComponent findFocused()
    {
        Window w = findTopWindow();
        if ( w==null )
            return null;

        JComponent c = null;
        Component c1 = w.getFocusOwner();
        while( true )
        {
            if ( c1 instanceof Scrollable
                || c1 instanceof JScrollPane )
            {
                c = (JComponent)c1;
                break;
            }
            else if ( c1 == null )
                break;
            c1 = c1.getParent();
        }

        return c;
    }


    /** very ugly + still needs a fix for titled borders */
    private JComponent findComponentAt( int x, int y )
    {
        Point p = new Point( x,y );
        Component c;
        try
        {
            //first: find the window
            Container cp = findTopWindow();
            trace("top window: " + cp);
            if ( cp==null )
                return null;
            Rectangle r = new Rectangle();
            Insets i = null;
            cp.getBounds( r );
            i = cp.getInsets();
            p.translate( -r.x-i.left, -r.y-i.top );
//            trace( "\ntop window: " + cp );

            //second: find window's content pane
            if ( cp instanceof JDialog )
                cp = ((JDialog)cp).getContentPane();
            else
                cp = ((JFrame)cp).getContentPane();
            cp.getBounds( r );
            i = cp.getInsets();
            p.translate( -r.x-i.left, -r.y-i.top );
//            trace( "\ncontent pane: " + cp );

            // now repeat the searching in the loop
            c = cp;
            Component c1 = null;
            while ( true )
            {
                if (c instanceof JTabbedPane)
                    c1 = ((JTabbedPane)c).getSelectedComponent();
                else if (c instanceof TabbedContentManager)
                    c1 = getComponentAt((Container)c, p);
                else
                    c1 = c.getComponentAt( p );
//illustration of JB4 TabbedContentManager bug
//                if (c1!=null)
//                {
//                    Rectangle rr = c1.getBounds();
//                    trace( "\ngoing down: "
//                            + c1.getClass().getName() + " "
//                            + rr.x + " " + rr.y + " "
//                            + rr.width + " " + rr.height);
//                }
//                else
//                    trace( "\ngoing down: NULL" );
//                if (c1 instanceof TabbedContentManager )
//                {
//                    TabbedContentManager tm = (TabbedContentManager)c1;
//                    //tm.get
//                    Component[] ccc = tm.getComponents();
//                    for (int xx=0; xx<ccc.length; xx++ )
//                        trace( "" + xx + ": " + ccc[xx].getClass().getName()
//                        + " " + ccc[xx].isVisible()
//                        + " " + ccc[xx].getBounds().width
//                        + " " + ccc[xx].getBounds().height );
//                }
                if ( c1 instanceof JScrollBar )
                    scrollFound = (JScrollBar)c1;
                if ( c1==c )
                    break;
                c = c1;
                if ( c1==null )
                    return null;
                c1.getBounds(r);
                i = ((Container)c).getInsets();
                p.translate( -r.x-i.left, -r.y-i.top );
            }

            // go up and return when something suitable for
            // scrolling found
            while( true )
            {
//                trace( "\ngoing up: " + (c1!=null ? c1.getClass().getName() :"NULL") );
                if ( c1 instanceof Scrollable
                    || c1 instanceof JScrollPane
                    || c1 instanceof JSlider
                    || c1 instanceof JComboBox )
                {
                    //System.err.println( "returning " + c1.getClass().getName() );
                    c = (JComponent)c1;
                    break;
                }
                else if ( c1 == null )
                    break;
                else if (c1 instanceof Container)
                {
                    Component[] allComp = ((Container)c1).getComponents();
                    JScrollBar bar = null;

                    for (int n = 0; n < allComp.length; n++)
                    {
                        if (allComp[n] instanceof JScrollBar)
                        {
                            bar = (JScrollBar)allComp[n];
                            int orientation = (action == Config.Action.hscroll || action == Config.Action.pghscroll)
                                        ? JScrollBar.HORIZONTAL
                                        : JScrollBar.VERTICAL;
                            if (bar.getOrientation() == orientation)
                            {
                                scrollFound = bar;
                                trace("JScrollbar (brother) - orientation OK");
                                return bar;
                            }
                        }
                    }
                    if (bar != null)
                    {
                        scrollFound = bar;
                        trace("JScrollbar (brother) - wrong orientation");
                        return bar;
                    }
                }
                c1 = c1.getParent();
            }
            return (JComponent)c;

        }
        catch ( ClassCastException ex )
        {
            ex.printStackTrace();
            return null;
        }
    }

    /** Helper function needed due to a bug(?) in TabbedContentManager
     *  in JB4: original getComponentAt function sometimes returns
     *  handle to an invisible component.
     *  @return  *
     */
    private Component getComponentAt(Container c, Point p)
    {
        Component children[] = c.getComponents();
        Component child;
        Rectangle compRect = new Rectangle();
        for (int i = 0; i < children.length; i++ )
        {
            child = children[i];
            if ( child.isVisible() )
            {
                child.getBounds(compRect);
                if (compRect.contains(p))
                    return child;
            }
        }
        return c;
    }

    /** ugly but works */
    private Window findTopWindow()
    {
        Browser b = Browser.getActiveBrowser();
        Component c = null;
        Frame[] fAll = b.getFrames();
        for (int i = 0; i < fAll.length; i++)
        {
//            trace("application frame: " + fAll[i].getTitle() + "; visible: " + fAll[i].isVisible() );

            Window w = findTopWindow(fAll[i]);//, "---");
            if (w != null)
                return w;

            c = fAll[i].getFocusOwner();
            if ( c!= null )
                return fAll[i];
        }

        return null;
    }

    private Window findTopWindow(Window w)//, String prefix)
    {
        Component c;
        Window[] wAll = w.getOwnedWindows();
        for ( int j = 0; j < wAll.length; j++ )
        {
//            trace(prefix + "window: " + wAll[j].getName() + "; visible: " + wAll[j].isVisible());
            // it looks like JBuilder 3.5 has a bug: even if dialog is
            // closed it's still on "owned windows" list.
            if ( wAll[j].isVisible() )
            {
                c = wAll[j].getFocusOwner();
                if ( c != null )
                    return wAll[j];
            }
            Window w1 = findTopWindow(wAll[j]);//, prefix+"---" );
            if (w1 != null)
                return w1;
        }
        return null;
    }


    /** Used if there is nothing better to scroll. Needed for Help
     windows
     */
    private void scrollScrollBar( int move )
    {
        int i = 0;
        if ( action==Config.Action.pgvscroll || action==Config.Action.pghscroll )
            i = scrollFound.getBlockIncrement( move>0 ? -1 : 1 );
        else
        {
            i = scrollFound.getUnitIncrement( move>0 ? -1 : 1 );
            i *= (conf.sensitivity*move/(NOTCH_MOVE*10));
        }
        i = scrollFound.getValue()-i;
        i = Math.min( i, scrollFound.getMaximum() );
        i = Math.max( i, scrollFound.getMinimum() );
        scrollFound.setValue( i );
    }

    private void scrollSlider( JSlider comp, int move )
    {
        //System.err.println( "scrollSlider" );
        int dv = (action==Config.Action.hscroll || action==Config.Action.vscroll)
            ? comp.getMinorTickSpacing() : comp.getMajorTickSpacing();
        if ( move<0 )
            dv = -dv;
        int v = comp.getValue() + dv;
        v = Math.min( v, comp.getMaximum() );
        v = Math.max( v, comp.getMinimum() );
        comp.setValue( v );
    }

    private void scrollCombo( JComboBox comp, int move )
    {
        //System.err.println( "scrollCombo" );
        //if ( comp.isPopupVisible() )
        //	return;
        int idx = comp.getSelectedIndex() + (move<0 ? 1 : -1);
        idx = Math.min( idx, comp.getItemCount()-1 );
        idx = Math.max( idx, 0 );
        comp.setSelectedIndex( idx );
    }

    protected void finalize()
                 throws Throwable
    {
        stopCaptureWheelMessages();
    }


    private final static MessageCategory WHEEL_MESSAGESS = new MessageCategory("wheel");

    private void trace( String msg )
    {
        if ( conf.debug )
        {
            System.out.println( msg );
         //   Browser.getActiveBrowser().getMessageView().addMessage(
         //               WHEEL_MESSAGESS,
         //               msg );
        }
    }

    /** An attempt to solve problem with freezing on JDK 1.3.
     */
    private class WheelNotifier implements Runnable
    {
        int move;
        int x;
        int y;
        int mask;

        long time;

        public void run()
        {
            wheelMoved( move, x, y, mask );
            move = 0;
        }
    }

}